@@ -119,7 +119,7 @@
                  (or #+x86-64 (%primitive !fast-alloc-immobile-symbol)
                      (alloc-immobile-fixedobj
                       symbol-size
-                      (logior (ash (1- symbol-size) n-widetag-bits) symbol-widetag))))))
+                      (compute-object-header (1- symbol-size) symbol-widetag))))))
     ;; symbol-hash and package ID start out as 0
     (%primitive set-slot symbol name 'make-symbol symbol-name-slot other-pointer-lowtag)
     (%primitive set-slot symbol nil 'make-symbol symbol-info-slot other-pointer-lowtag)
@@ -288,6 +288,20 @@
     t))
 
 #+permgen
+(progn
+(defun allocate-permgen-symbol (name)
+  (with-system-mutex (*allocator-mutex* :without-gcing t)
+    (let ((freeptr *permgen-space-free-pointer*))
+      (setf *permgen-space-free-pointer*
+            (sap+ freeptr (ash symbol-size word-shift)))
+      (aver (<= (sap-int *permgen-space-free-pointer*)
+                (+ permgen-space-start permgen-space-size)))
+      (setf (sap-ref-word freeptr 0) symbol-widetag)
+      (setf (sap-ref-lispobj freeptr (ash symbol-name-slot word-shift)) name
+            (sap-ref-lispobj freeptr (ash symbol-info-slot word-shift)) nil
+            (sap-ref-word freeptr (ash symbol-value-slot word-shift))
+            unbound-marker-widetag)
+      (%make-lisp-obj (sap-int (sap+ freeptr other-pointer-lowtag))))))
 (defun sb-kernel::allocate-permgen-layout (nwords)
   (with-system-mutex (*allocator-mutex* :without-gcing t)
     (let ((freeptr *permgen-space-free-pointer*))
@@ -298,4 +312,4 @@
                 (+ permgen-space-start permgen-space-size)))
       (setf (sap-ref-word freeptr 0)
             (logior (ash nwords instance-length-shift) instance-widetag))
-      (%make-lisp-obj (sap-int (sap+ freeptr instance-pointer-lowtag))))))
+      (%make-lisp-obj (sap-int (sap+ freeptr instance-pointer-lowtag)))))))
