@@ -614,136 +614,384 @@
   (def :cr define-utf8->string/cr utf8->string/cr bytes-per-utf8-character simple-get-utf8-char/cr)
   (def :crlf define-utf8-string/crlf utf8->string/crlf bytes-per-utf8-character/crlf simple-get-utf8-char/crlf))
 
-#+(and sb-unicode 64-bit little-endian)
-(defun copy-utf8-bytes-to-character-string (requested total-copied start buffer ibuf)
-  (declare (type index start requested total-copied)
+#+(and sb-unicode 64-bit little-endian
+       (not (or arm64 x86-64))) ;; have true simd definitions
+(defun sb-vm::simd-copy-utf8-to-character-string (start end string ibuf)
+  (declare (type index start end)
            (optimize speed (safety 0)))
-  (with-pinned-objects (buffer)
+  (with-pinned-objects (string)
     (let* ((head (buffer-head ibuf))
            (tail (buffer-tail ibuf))
            (sap (buffer-sap ibuf))
-           (left (- requested total-copied))
+           (left (- end start))
            (result-characters (logand left (- sb-vm:n-word-bytes)))
-           (buffer-bytes (logand (- tail head) (- sb-vm:n-word-bytes)))
-           (n (min result-characters buffer-bytes))
+           (string-bytes (logand (- tail head) (- sb-vm:n-word-bytes)))
+           (n (min result-characters string-bytes))
            (repeat (ldb (byte sb-vm:n-word-bits 0) #x0101010101010101))
            (ascii-mask (* 128 repeat))
-           (buffer-sap (vector-sap buffer))
-           (buffer-start (truly-the fixnum (* (+ start total-copied) 4))))
+           (string-sap (vector-sap string))
+           (string-offset (truly-the fixnum (* start 4))))
       (declare (optimize sb-c::preserve-single-use-debug-variables
                          sb-c::preserve-constants))
       (loop for ibuf-offset from head below (+ head n) by sb-vm:n-word-bytes
-            for buffer-offset from buffer-start by (* 4 sb-vm:n-word-bytes)
             do
-            (let ((word (sb-sys:sap-ref-word sap ibuf-offset)))
+            (let ((word (sap-ref-word sap ibuf-offset)))
               (when (logtest word ascii-mask)
                 (setf head ibuf-offset)
-                (incf total-copied
-                      (truncate (- buffer-offset buffer-start) 4))
                 (return))
-              (setf (sap-ref-word buffer-sap buffer-offset)
+              (setf (sap-ref-word string-sap string-offset)
                     (dpb (ldb (byte 8 8) word)
                          (byte 8 32)
                          (ldb (byte 8 0) word))
-                    (sap-ref-word buffer-sap (+ buffer-offset 8))
+                    (sap-ref-word string-sap (+ string-offset 8))
                     (dpb (ldb (byte 8 24) word)
                          (byte 8 32)
                          (ldb (byte 8 16) word))
-                    (sap-ref-word buffer-sap (+ buffer-offset 16))
+                    (sap-ref-word string-sap (+ string-offset 16))
                     (dpb (ldb (byte 8 40) word)
                          (byte 8 32)
                          (ldb (byte 8 32) word))
-                    (sap-ref-word buffer-sap (+ buffer-offset 24))
+                    (sap-ref-word string-sap (+ string-offset 24))
                     (dpb (ldb (byte 8 56) word)
                          (byte 8 32)
                          (ldb (byte 8 48) word))))
-            finally (incf head n)
-                    (incf total-copied n))
+            (incf string-offset (* 4 sb-vm:n-word-bytes))
+            finally (incf head n))
       (setf (buffer-head ibuf) head)
-      total-copied)))
+      (truly-the index (truncate string-offset 4)))))
 
-(defun fd-stream-read-n-characters/utf-8 (stream buffer sbuffer start requested eof-error-p &aux (total-copied 0))
-  (declare (type fd-stream stream)
-           (type index start requested total-copied)
-           (type ansi-stream-cin-buffer buffer)
-           (type ansi-stream-csize-buffer sbuffer))
-  (when (fd-stream-eof-forced-p stream)
-    (setf (fd-stream-eof-forced-p stream) nil)
-    (return-from fd-stream-read-n-characters/utf-8 0))
-  (do ((instead (fd-stream-instead stream))
-       (index (+ start total-copied) (1+ index)))
-      ((= (fill-pointer instead) 0)
-       (setf (fd-stream-listen stream) nil))
-    (setf (aref buffer index) (vector-pop instead))
-    (setf (aref sbuffer index) 0)
-    (incf total-copied)
-    (when (= requested total-copied)
-      (when (= (fill-pointer instead) 0)
-        (setf (fd-stream-listen stream) nil))
-      (return-from fd-stream-read-n-characters/utf-8 total-copied)))
-  #+(and sb-unicode 64-bit little-endian)
-  (let ((new-total (copy-utf8-bytes-to-character-string requested total-copied
-                                                        start buffer (fd-stream-ibuf stream))))
-    (fill sbuffer 1 :start (+ start total-copied) :end (+ start new-total))
-    (setf total-copied new-total))
-  (do ()
-      (())
-    (let* ((ibuf (fd-stream-ibuf stream))
-           (head (buffer-head ibuf))
+#+(and sb-unicode 64-bit little-endian)
+(defun sb-vm::simd-copy-utf8-crlf-to-character-string-with-size (start end string ibuf size-buffer)
+  (declare (type index start end)
+           (optimize speed (safety 0)))
+  (with-pinned-objects (string size-buffer)
+    (let* ((head (buffer-head ibuf))
            (tail (buffer-tail ibuf))
            (sap (buffer-sap ibuf))
-           (decode-break-reason nil))
-      (declare (type index head tail))
-      (do ((size nil nil))
-          ((or (= tail head) (= requested total-copied)))
-        (setf decode-break-reason
-              (block decode-break-reason
-                (when (< (- tail head) 1) (return))
-                (let ((byte (sap-ref-8 sap head)))
-                  (declare (ignorable byte))
-                  (setq size (cond ((< byte 128) 1) ((< byte 194) (return-from decode-break-reason 1)) ((< byte 224) 2) ((< byte 240) 3) (t 4)))
-                  (when (> size (- tail head)) (return))
-                  (let ((index (+ start total-copied)))
-                    (setf (aref buffer index)
-                          (code-char
-                           (ecase size
-                             (1 byte)
-                             (2
-                              (let ((byte2 (sap-ref-8 sap (1+ head))))
-                                (unless (<= 128 byte2 191) (return-from decode-break-reason 2))
-                                (dpb byte (byte 5 6) byte2)))
-                             (3
-                              (let ((byte2 (sap-ref-8 sap (1+ head))) (byte3 (sap-ref-8 sap (+ 2 head))))
-                                (unless
-                                    (and (<= 128 byte2 191) (<= 128 byte3 191) (or (/= byte 224) (<= 160 byte2 191))
-                                         (or (/= byte 237) (<= 128 byte2 159)))
-                                  (return-from decode-break-reason 3))
-                                (dpb byte (byte 4 12) (dpb byte2 (byte 6 6) byte3))))
-                             (4
-                              (let ((byte2 (sap-ref-8 sap (1+ head))) (byte3 (sap-ref-8 sap (+ 2 head))) (byte4 (sap-ref-8 sap (+ 3 head))))
-                                (unless
-                                    (and (<= 128 byte2 191) (<= 128 byte3 191) (<= 128 byte4 191) (or (/= byte 240) (<= 144 byte2 191))
-                                         (or (/= byte 244) (<= 128 byte2 143)))
-                                  (return-from decode-break-reason 4))
-                                (dpb byte (byte 3 18) (dpb byte2 (byte 6 12) (dpb byte3 (byte 6 6) byte4))))))))
-                    (setf (aref sbuffer index) size))
-                  (incf total-copied)
-                  (incf head size))
-                nil))
-        (setf (buffer-head ibuf) head)
-        (when decode-break-reason
-          (when (plusp total-copied) (return-from fd-stream-read-n-characters/utf-8 total-copied))
-          (when (stream-decoding-error-and-handle stream decode-break-reason 1)
-            (if eof-error-p
-                (error 'end-of-file :stream stream)
-                (return-from fd-stream-read-n-characters/utf-8 total-copied)))
-          (return-from fd-stream-read-n-characters/utf-8 total-copied)))
+           (n (logand (min (- end start)
+                           (- (- tail head) 1))
+                      (- 2)))
+           (repeat (ldb (byte 16 0) #x0101010101010101))
+           (ascii-mask (* 128 repeat))
+           (string-sap (vector-sap string))
+           (size-sap (vector-sap size-buffer))
+           (size-offset start)
+           (string-offset (* start 4))
+           (end (+ head n)))
+      (declare (index string-offset size-offset))
+      (declare (optimize sb-c::preserve-single-use-debug-variables
+                         sb-c::preserve-constants))
+      (loop while (< head end)
+            do
+            (let ((word (sap-ref-16 sap head)))
+              (when (logtest word ascii-mask)
+                (return))
+              (cond ((= word #x0A0D)
+                     (setf (sap-ref-32 string-sap string-offset) 10
+                           (sap-ref-8 size-sap size-offset) 2)
+                     (incf head 2)
+                     (incf string-offset 4)
+                     (incf size-offset 1))
+                    ((= (ash word -8) 13)
+                     (setf (sap-ref-32 string-sap string-offset) (ldb (byte 8 0) word)
+                           (sap-ref-8 size-sap size-offset) 1)
+                     (incf head 1)
+                     (incf string-offset 4)
+                     (incf size-offset 1))
+                    (t
+                     (setf (sap-ref-64 string-sap string-offset)
+                           (dpb (ldb (byte 8 8) word)
+                                (byte 32 32)
+                                (ldb (byte 8 0) word))
+                           (sap-ref-16 size-sap size-offset) #x0101)
+                     (incf head 2)
+                     (incf string-offset 8)
+                     (incf size-offset 2)))))
       (setf (buffer-head ibuf) head)
-      (cond ((plusp total-copied) (return total-copied))
-            ((or (eq decode-break-reason 'eof) (null (catch 'eof-input-catcher (refill-input-buffer stream))))
-             (if eof-error-p
-                 (error 'end-of-file :stream stream)
-                 (return total-copied)))))))
+      size-offset)))
+
+#+(and sb-unicode 64-bit little-endian
+       (not (or arm64 x86-64))) ;; have true simd definitions
+(defun sb-vm::simd-copy-utf8-to-base-string (start end string ibuf)
+  (declare (type index start end)
+           (optimize speed (safety 0)))
+  (with-pinned-objects (string)
+    (let* ((head (buffer-head ibuf))
+           (tail (buffer-tail ibuf))
+           (sap (buffer-sap ibuf))
+           (n (logand (min (- (- end start))
+                           (- tail head))
+                      (- sb-vm:n-word-bytes)))
+           (repeat (ldb (byte sb-vm:n-word-bits 0) #x0101010101010101))
+           (ascii-mask (* 128 repeat))
+           (string-sap (vector-sap string))
+           (string-offset start))
+      (declare (optimize sb-c::preserve-single-use-debug-variables
+                         sb-c::preserve-constants))
+      (loop for ibuf-offset from head below (+ head n) by sb-vm:n-word-bytes
+            do
+            (incf string-offset sb-vm:n-word-bytes)
+            (let ((word (sap-ref-word sap ibuf-offset)))
+              (when (logtest word ascii-mask)
+                (setf head ibuf-offset)
+                (return))
+              (setf (sap-ref-word string-sap string-offset) word))
+            finally (incf head n))
+      (setf (buffer-head ibuf) head)
+      (truly-the index string-offset))))
+
+(defmacro utf8-char-loop (&key size-buffer
+                               crlf
+                               (eof t))
+  `(do ()
+       ((or (= tail head)
+            (= index end)))
+     (let ((byte (sap-ref-8 sap head)))
+       (setf (aref string index)
+             (code-char
+              (cond
+                ,@(and crlf
+                       `(((and (= byte 13)
+                               (let ((new-head (+ head 2)))
+                                 (cond ((and (<= new-head tail)
+                                             (= (sap-ref-8 sap (1+ head)) 10))
+                                        (incf head 2)
+                                        ,@(when size-buffer
+                                            `((setf (aref size-buffer index) 2)))
+                                        10)
+                                       ,@(and eof
+                                          `((eof nil)))
+                                       (t
+                                        ,@(and eof
+                                               `((setf requested-refill t)))
+                                        (return))))))))
+                ((< byte 128)
+                 (incf head)
+                 ,@(when size-buffer
+                     `((setf (aref size-buffer index) 1)))
+                 byte)
+                ((< byte 194)
+                 (decode-break 1))
+                ((< byte 224)
+                 (let ((new-head (+ head 2)))
+                   (when (> new-head tail)
+                     ,@(and eof
+                            `((setf incomplete t)))
+                     (return))
+                   (prog1
+                       (let ((byte2 (sap-ref-8 sap (1+ head))))
+                         (unless (<= 128 byte2 191)
+                           (decode-break 2))
+                         (dpb byte (byte 5 6) byte2))
+                     ,@(when size-buffer
+                         `((setf (aref size-buffer index) 2)))
+                     (setf head new-head))))
+                ((< byte 240)
+                 (let ((new-head (+ head 3)))
+                   (when (> new-head tail)
+                     ,@(and eof
+                            `((setf incomplete t)))
+                     (return))
+                   (prog1
+                       (let ((byte2 (sap-ref-8 sap (1+ head)))
+                             (byte3 (sap-ref-8 sap (+ 2 head))))
+                         (unless
+                             (and (<= 128 byte2 191)
+                                  (<= 128 byte3 191)
+                                  (or (/= byte 224) (<= 160 byte2 191))
+                                  (or (/= byte 237) (<= 128 byte2 159)))
+                           (decode-break 3))
+                         (dpb byte (byte 4 12)
+                              (dpb byte2 (byte 6 6) byte3)))
+                     ,@(when size-buffer
+                         `((setf (aref size-buffer index) 3)))
+                     (setf head new-head))))
+                (t
+                 (let ((new-head (+ head 4)))
+                   (when (> new-head tail)
+                     ,@(and eof
+                            `((setf incomplete t)))
+                     (return))
+                   (prog1
+                       (let ((byte2 (sap-ref-8 sap (1+ head)))
+                             (byte3 (sap-ref-8 sap (+ 2 head)))
+                             (byte4 (sap-ref-8 sap (+ 3 head))))
+                         (unless
+                             (and (<= 128 byte2 191)
+                                  (<= 128 byte3 191)
+                                  (<= 128 byte4 191)
+                                  (or (/= byte 240) (<= 144 byte2 191))
+                                  (or (/= byte 244) (<= 128 byte2 143)))
+                           (decode-break 4))
+                         (dpb byte (byte 3 18)
+                              (dpb byte2 (byte 6 12)
+                                   (dpb byte3 (byte 6 6) byte4))))
+                     ,@(when size-buffer
+                         `((setf (aref size-buffer index) 4)))
+                     (setf head new-head)))))))
+       (incf index))
+     (setf (buffer-head ibuf) head)))
+
+(defun fd-stream-read-n-characters/utf-8 (stream string size-buffer start end &aux (index start))
+  (declare (type fd-stream stream)
+           (type index start end index)
+           (type ansi-stream-cin-buffer string)
+           (type ansi-stream-csize-buffer size-buffer))
+  (when (fd-stream-eof-forced-p stream)
+    (setf (fd-stream-eof-forced-p stream) nil)
+    (return-from fd-stream-read-n-characters/utf-8 start))
+  (do ((instead (fd-stream-instead stream)))
+      ((= (fill-pointer instead) 0)
+       (setf (fd-stream-listen stream) nil))
+    (setf (aref string index) (vector-pop instead))
+    (setf (aref size-buffer index) 0)
+    (incf index)
+    (when (= end index)
+      (when (= (fill-pointer instead) 0)
+        (setf (fd-stream-listen stream) nil))
+      (return-from fd-stream-read-n-characters/utf-8 index)))
+  (block outer
+    (do ()
+        (())
+      #+(and sb-unicode 64-bit little-endian)
+      (let ((new-index (sb-vm::simd-copy-utf8-to-character-string index end string (fd-stream-ibuf stream))))
+        ;; Make sure to change this 1 whenever
+        ;; simd-copy-utf8-to-character-string starts processing more than
+        ;; just ascii characters.
+        (fill size-buffer 1 :start index :end new-index)
+        (setf index new-index))
+      (let* ((ibuf (fd-stream-ibuf stream))
+             (head (buffer-head ibuf))
+             (tail (buffer-tail ibuf))
+             (sap (buffer-sap ibuf)))
+        (declare (type index head tail))
+        (flet ((decode-break (reason)
+                 (setf (buffer-head ibuf) head)
+                 (unless (> index start)
+                   (stream-decoding-error-and-handle stream reason 1))
+                 (return-from outer index)))
+          (utf8-char-loop :size-buffer t
+                          :eof nil))
+        (when (or (> index start)
+                  (null (catch 'eof-input-catcher (refill-input-buffer stream))))
+          (return index))))))
+
+(defun fd-stream-read-n-characters/utf-8/crlf (stream string size-buffer start end &aux (index start))
+  (declare (type fd-stream stream)
+           (type index start end index)
+           (type ansi-stream-cin-buffer string)
+           (type ansi-stream-csize-buffer size-buffer))
+  (when (fd-stream-eof-forced-p stream)
+    (setf (fd-stream-eof-forced-p stream) nil)
+    (return-from fd-stream-read-n-characters/utf-8/crlf start))
+  (do ((instead (fd-stream-instead stream)))
+      ((= (fill-pointer instead) 0)
+       (setf (fd-stream-listen stream) nil))
+    (setf (aref string index) (vector-pop instead))
+    (setf (aref size-buffer index) 0)
+    (incf index)
+    (when (= end index)
+      (when (= (fill-pointer instead) 0)
+        (setf (fd-stream-listen stream) nil))
+      (return-from fd-stream-read-n-characters/utf-8/crlf index)))
+  (block outer
+    (do ()
+        (())
+      #+(and sb-unicode 64-bit little-endian)
+      (setf index
+            (sb-vm::simd-copy-utf8-crlf-to-character-string-with-size index end string (fd-stream-ibuf stream) size-buffer))
+      (let* ((ibuf (fd-stream-ibuf stream))
+             (head (buffer-head ibuf))
+             (tail (buffer-tail ibuf))
+             (sap (buffer-sap ibuf)))
+        (declare (type index head tail))
+        (flet ((decode-break (reason)
+                 (setf (buffer-head ibuf) head)
+                 (unless (> index start)
+                   (stream-decoding-error-and-handle stream reason 1))
+                 (return-from outer index)))
+          (utf8-char-loop :size-buffer t :crlf t :eof nil))
+        (when (or (> index start)
+                  (null (catch 'eof-input-catcher (refill-input-buffer stream))))
+          (return index))))))
+
+;;; Bypass the character buffer, the caller must ensure that it's empty
+(defun fd-stream-read-sequence/utf-8-to-string (stream string start end &aux (index start))
+  (declare (type fd-stream stream)
+           (type index start end index)
+           (type (simple-array character (*)) string))
+  (block outer
+    (loop
+     (do ((instead (fd-stream-instead stream)))
+         ((= (fill-pointer instead) 0)
+          (setf (fd-stream-listen stream) nil))
+       (setf (aref string index) (vector-pop instead))
+       (incf index)
+       (when (= index end)
+         (when (= (fill-pointer instead) 0)
+           (setf (fd-stream-listen stream) nil))
+         (return index)))
+     #+(and sb-unicode 64-bit little-endian)
+     (setf index
+           (sb-vm::simd-copy-utf8-to-character-string index end string (fd-stream-ibuf stream)))
+     (let* ((ibuf (fd-stream-ibuf stream))
+            (head (buffer-head ibuf))
+            (tail (buffer-tail ibuf))
+            (sap (buffer-sap ibuf))
+            incomplete)
+       (declare (type index head tail))
+       (flet ((decode-break (reason)
+                (setf (buffer-head ibuf) head)
+                (when (stream-decoding-error-and-handle stream reason 1)
+                  (return-from outer index))
+                (return)))
+         (utf8-char-loop))
+       (when (= index end)
+         (setf (buffer-head ibuf) head)
+         (return index))
+       (unless (catch 'eof-input-catcher (refill-input-buffer stream))
+         (when (or (not incomplete)
+                   (stream-decoding-error-and-handle stream (- tail head) 1))
+           (return index)))))))
+
+#+sb-unicode
+(defun fd-stream-read-sequence/utf-8-to-base-string (stream string start end &aux (index start))
+  (declare (type fd-stream stream)
+           (type index start end index)
+           (type simple-base-string string))
+  (block outer
+    (loop
+     (do ((instead (fd-stream-instead stream)))
+         ((= (fill-pointer instead) 0)
+          (setf (fd-stream-listen stream) nil))
+       (setf (aref string index) (vector-pop instead))
+       (incf index)
+       (when (= index end)
+         (when (= (fill-pointer instead) 0)
+           (setf (fd-stream-listen stream) nil))
+         (return index)))
+     #+(and sb-unicode 64-bit little-endian)
+     (setf index
+           (sb-vm::simd-copy-utf8-to-base-string index end string (fd-stream-ibuf stream)))
+     (let* ((ibuf (fd-stream-ibuf stream))
+            (head (buffer-head ibuf))
+            (tail (buffer-tail ibuf))
+            (sap (buffer-sap ibuf))
+            incomplete)
+       (declare (type index head tail))
+       (flet ((decode-break (reason)
+                (setf (buffer-head ibuf) head)
+                (when (stream-decoding-error-and-handle stream reason 1)
+                  (return-from outer index))
+                (return)))
+         (utf8-char-loop))
+       (when (= index end)
+         (setf (buffer-head ibuf) head)
+         (return index))
+       (unless (catch 'eof-input-catcher (refill-input-buffer stream))
+         (when (or (not incomplete)
+                   (stream-decoding-error-and-handle stream (- tail head) 1))
+           (return index)))))))
 
 (define-external-format/variable-width (:utf-8 :utf8) t
   #+sb-unicode (code-char #xfffd) #-sb-unicode #\?
@@ -883,9 +1131,12 @@
              (sap-ref-8 sap (+ 3 tail)) (logior #x80 (ldb (byte 6 0) bits)))))
   ((2 1)
    (cond
-     ((= (- tail head) 1) 1)
-     ((and (= byte 13) (= (sap-ref-8 sap (+ head 1)) 10)) 2)
-     ((< byte #x80) 1)
+     ((< byte #x80)
+      (if (and (= byte 13)
+               (> (- tail head) 1)
+               (= (sap-ref-8 sap (+ head 1)) 10))
+          2
+          1))
      ((< byte #xc2) (return-from decode-break-reason 1))
      ((< byte #xe0) 2)
      ((< byte #xf0) 3)
@@ -921,4 +1172,177 @@
   utf8->string/crlf-aref
   string/crlf->utf8
   :char-encodable-p (let ((bits (char-code |ch|))) (not (<= #xd800 bits #xdfff)))
+  :fd-stream-read-n-characters fd-stream-read-n-characters/utf-8/crlf
   :newline-variant :crlf)
+
+#+(and sb-unicode 64-bit little-endian
+       (not arm64)) ;; have true simd definitions or might be redefined with def-variant
+(defun sb-vm::simd-copy-utf8-crlf-to-base-string (start end string ibuf)
+  (declare (type index start end)
+           (optimize speed (safety 0)))
+  (with-pinned-objects (string)
+    (let* ((head (buffer-head ibuf))
+           (tail (buffer-tail ibuf))
+           (sap (buffer-sap ibuf))
+           (n (logand (min (- end start)
+                           (- (- tail head) 1))
+                      (- 2)))
+           (repeat (ldb (byte 16 0) #x0101010101010101))
+           (ascii-mask (* 128 repeat))
+           (string-sap (vector-sap string))
+           (string-start start)
+           (string-offset string-start)
+           (end (+ head n)))
+      (declare (optimize sb-c::preserve-single-use-debug-variables
+                         sb-c::preserve-constants))
+      (loop while (< head end)
+            do
+            (let ((word (sap-ref-16 sap head)))
+              (when (logtest word ascii-mask)
+                (return))
+              (cond ((= word #x0A0D)
+                     (setf (sap-ref-8 string-sap string-offset) 10)
+                     (incf head 2)
+                     (incf string-offset 1))
+                    ((= (ash word -8) 13)
+                     (setf (sap-ref-8 string-sap string-offset) word)
+                     (incf head 1)
+                     (incf string-offset 1))
+                    (t
+                     (setf (sap-ref-16 string-sap string-offset) word)
+                     (incf head 2)
+                     (incf string-offset 2)))))
+      (setf (buffer-head ibuf) head)
+      (truly-the index string-offset))))
+
+#+(and sb-unicode 64-bit little-endian
+       (not arm64)) ;; have true simd definitions or might be redefined with def-variant
+(defun sb-vm::simd-copy-utf8-crlf-to-character-string (start end string ibuf)
+  (declare (type index start end)
+           (optimize speed (safety 0)))
+  (with-pinned-objects (string)
+    (let* ((head (buffer-head ibuf))
+           (tail (buffer-tail ibuf))
+           (sap (buffer-sap ibuf))
+           (n (logand (min (- end start)
+                           (- (- tail head) 1))
+                      (- 2)))
+           (repeat (ldb (byte 16 0) #x0101010101010101))
+           (ascii-mask (* 128 repeat))
+           (string-sap (vector-sap string))
+           (string-offset (* start 4))
+           (end (+ head n)))
+      (declare (optimize sb-c::preserve-single-use-debug-variables
+                         sb-c::preserve-constants))
+      (loop while (< head end)
+            do
+            (let ((word (sap-ref-16 sap head)))
+              (when (logtest word ascii-mask)
+                (return))
+              (cond ((= word #x0A0D)
+                     (setf (sap-ref-32 string-sap string-offset) 10)
+                     (incf head 2)
+                     (incf string-offset 4))
+                    ((= (ash word -8) 13)
+                     (setf (sap-ref-32 string-sap string-offset) (ldb (byte 8 0) word))
+                     (incf head 1)
+                     (incf string-offset 4))
+                    (t
+                     (setf (sap-ref-64 string-sap string-offset)
+                           (dpb (ldb (byte 8 8) word)
+                                (byte 32 32)
+                                (ldb (byte 8 0) word)))
+                     (incf head 2)
+                     (incf string-offset 8)))))
+      (setf (buffer-head ibuf) head)
+      (truly-the index (truncate string-offset 4)))))
+
+(defun fd-stream-read-sequence/utf-8-crlf-to-character-string (stream string start end)
+  (declare (type fd-stream stream)
+           (type index start end)
+           (type (simple-array character (*)) string))
+  (let (eof
+        (index start))
+    (declare (index index))
+    (block outer
+      (loop
+       (do ((instead (fd-stream-instead stream)))
+           ((= (fill-pointer instead) 0)
+            (setf (fd-stream-listen stream) nil))
+         (setf (aref string index) (vector-pop instead))
+         (incf index)
+         (when (= index end)
+           (when (= (fill-pointer instead) 0)
+             (setf (fd-stream-listen stream) nil))
+           (return index)))
+       #+(and sb-unicode 64-bit little-endian)
+       (setf index
+             (sb-vm::simd-copy-utf8-crlf-to-character-string index end string (fd-stream-ibuf stream)))
+       (let* ((ibuf (fd-stream-ibuf stream))
+              (head (buffer-head ibuf))
+              (tail (buffer-tail ibuf))
+              (sap (buffer-sap ibuf))
+              requested-refill
+              incomplete)
+         (declare (type index head tail))
+         (flet ((decode-break (reason)
+                  (setf (buffer-head ibuf) head)
+                  (when (stream-decoding-error-and-handle stream reason 1)
+                    (return-from outer index))
+                  (return)))
+           (utf8-char-loop :crlf t))
+         (when (or (= end index)
+                   eof)
+           (setf (buffer-head ibuf) head)
+           (return index))
+         (unless (catch 'eof-input-catcher (refill-input-buffer stream))
+           (if requested-refill
+               (setf eof t)
+               (when (or (not incomplete)
+                         (stream-decoding-error-and-handle stream (- tail head) 1))
+                 (return index)))))))))
+
+(defun fd-stream-read-sequence/utf-8-crlf-to-base-string (stream string start end)
+  (declare (type fd-stream stream)
+           (type index start end)
+           (type simple-base-string string))
+  (let (eof
+        (index start))
+    (declare (index index))
+    (block outer
+      (loop
+       (do ((instead (fd-stream-instead stream)))
+           ((= (fill-pointer instead) 0)
+            (setf (fd-stream-listen stream) nil))
+         (setf (aref string index) (vector-pop instead))
+         (incf index)
+         (when (= index end)
+           (when (= (fill-pointer instead) 0)
+             (setf (fd-stream-listen stream) nil))
+           (return index)))
+       #+(and sb-unicode 64-bit little-endian)
+       (setf index
+             (sb-vm::simd-copy-utf8-crlf-to-base-string index end string (fd-stream-ibuf stream)))
+       (let* ((ibuf (fd-stream-ibuf stream))
+              (head (buffer-head ibuf))
+              (tail (buffer-tail ibuf))
+              (sap (buffer-sap ibuf))
+              requested-refill
+              incomplete)
+         (declare (type index head tail))
+         (flet ((decode-break (reason)
+                  (setf (buffer-head ibuf) head)
+                  (when (stream-decoding-error-and-handle stream reason 1)
+                    (return-from outer index))
+                  (return)))
+           (utf8-char-loop :crlf t))
+         (when (or (= end index)
+                   eof)
+           (setf (buffer-head ibuf) head)
+           (return index))
+         (unless (catch 'eof-input-catcher (refill-input-buffer stream))
+           (if requested-refill
+               (setf eof t)
+               (when (or (not incomplete)
+                         (stream-decoding-error-and-handle stream (- tail head) 1))
+                 (return index)))))))))
