@@ -227,6 +227,7 @@
            (let ((*print-readably* nil)
                  (*current-level-in-print* 0)
                  (*standard-output* (make-string-output-stream))
+                 (*print-pretty* t)
                  (*in-trace* t))
              (case (trace-info-report info)
                (trace
@@ -538,7 +539,7 @@
             (error "unknown TRACE option: ~S" name))
            ((stringp name)
             (let ((package (find-undeleted-package-or-lose name)))
-              (do-all-symbols (symbol (find-package name))
+              (do-symbols (symbol package)
                 (when (eql package (symbol-package symbol))
                   (when (and (fboundp symbol)
                              (not (macro-function symbol))
@@ -803,14 +804,14 @@ functions when called with no arguments."
                                    ;; The code constants will be overwritten in the copy.
                                    ;; These are just placeholders essentially.
                                    (apply ,#'trace-call ,(make-trace-info) #() args))))))
-                 (index (+ sb-vm:code-constants-offset sb-vm:code-slots-per-simple-fun)))
+                 (index sb-vm:code-constants-offset))
              ;; First three args to APPLY must be at the expected offets
              (aver (typep (code-header-ref c (+ index 0)) 'function))
              (aver (typep (code-header-ref c (+ index 1)) 'trace-info))
              (aver (typep (code-header-ref c (+ index 2)) 'simple-vector))
              c)
            t)))
-        (index (+ sb-vm:code-constants-offset sb-vm:code-slots-per-simple-fun)))
+        (index sb-vm:code-constants-offset))
     (setf (code-header-ref code (+ index 0)) (symbol-function wrapper)
           (code-header-ref code (+ index 1)) info
           (code-header-ref code (+ index 2)) actual-fun)
@@ -882,5 +883,5 @@ functions when called with no arguments."
                    tracing-wrapper))))))
     ;; Possibly update #'NAME to point to the tracing wrapper
     (when (and namep (eq (fboundp name) traced-fun))
-      (setf (fdefn-fun (find-fdefn name)) tracing-wrapper))
+      (fset name tracing-wrapper))
     tracing-wrapper))
