@@ -69,6 +69,7 @@
 (export 'sb-kernel::*save-lisp-clobbered-globals* 'sb-kernel)
 (define-load-time-global sb-kernel::*save-lisp-clobbered-globals*
     '#(sb-impl::*exit-lock*
+       sb-vm::*allocator-mutex*
        sb-thread::*make-thread-lock*
        sb-thread::*initial-thread*
        ;; Saving *JOINABLE-THREADS* could cause catastophic failure on restart.
@@ -250,6 +251,7 @@ sufficiently motivated to do lengthy fixes."
           ;; A normal GC will leave huge amounts of storage unreclaimed
           ;; (over 50% on x86). This needs to be done by a single function
           ;; since the GC will invalidate the stack.
+          (sb-kernel::unsafe-clear-roots sb-vm:+highest-normal-generation+)
           (gc-and-save name
                        (foreign-bool executable)
                        (foreign-bool purify)
@@ -336,15 +338,12 @@ sufficiently motivated to do lengthy fixes."
     ;; recreate it so that we don't preserve an empty vector taking up 16KB
     (setq sb-kernel::*forward-referenced-layouts* (make-hash-table :test 'equal)))
   ;; Clean up the simulated weak list of covered code components.
-  (rplacd sb-c:*code-coverage-info*
-          (delete-if-not #'weak-pointer-value (cdr sb-c:*code-coverage-info*)))
+  (rplacd *code-coverage-info*
+          (delete-if-not #'weak-pointer-value (cdr *code-coverage-info*)))
   (sb-kernel::rebuild-ctype-hashsets)
   (drop-all-hash-caches)
   (os-deinit)
   (clrhash sb-c::*emitted-full-calls*) ; Don't immortalize compiler's scratchpad
-  ;; Perform static linkage. Functions become un-statically-linked
-  ;; on demand, for TRACE, redefinition, etc.
-  #+(and immobile-code x86-64) (sb-vm::statically-link-core)
   (finalizers-deinit)
   ;; Try to shrink the pathname cache. It might be largely nulls
   (rebuild-pathname-cache)
@@ -360,16 +359,18 @@ sufficiently motivated to do lengthy fixes."
 (in-package "SB-C")
 
 (defun coalesce-debug-info ()
+  ;; Discard the uncompacted fun map cache.
+  (setq sb-di::*uncompacted-fun-maps* nil)
+  ;; Discard the debugger's cached mapping of debug functions.
+  (setq sb-di::*compiled-debug-funs* nil)
   (flet ((debug-source= (a b)
            (and (equalp a b)
                 ;; Case sensitive
                 (equal (debug-source-plist a) (debug-source-plist b)))))
     ;; Coalesce the following:
-    ;;  DEBUG-INFO-SOURCE, DEBUG-FUN-NAME
-    ;;  SIMPLE-FUN-ARGLIST, SIMPLE-FUN-TYPE
+    ;;  DEBUG-INFO-SOURCE, SIMPLE-FUN-ARGLIST, SIMPLE-FUN-TYPE
     ;; FUN-NAMES-EQUALISH considers any two string= gensyms as EQ.
     (let ((source-ht (make-hash-table :test 'equal))
-          (name-ht (make-hash-table :test 'equal))
           (arglist-hash (make-hash-table :hash-function 'sb-impl::equal-hash
                                          :test 'sb-impl::fun-names-equalish))
           (type-hash (make-hash-table :test 'equal)))
@@ -378,10 +379,6 @@ sufficiently motivated to do lengthy fixes."
          (declare (ignore size))
          (case widetag
            (#.sb-vm:code-header-widetag
-            (let ((di (%code-debug-info obj)))
-              ;; Discard memoized debugger's debug info
-              (when (typep di 'sb-c::compiled-debug-info)
-                (setf (sb-c::compiled-debug-info-memo-cell di) nil)))
             (dotimes (i (sb-kernel:code-n-entries obj))
               (let* ((fun (sb-kernel:%code-entry-point obj i))
                      (arglist (%simple-fun-arglist fun))
@@ -410,18 +407,7 @@ sufficiently motivated to do lengthy fixes."
                              (push source (gethash namestring source-ht)))
                             ((neq source canonical-repr)
                              (setf (compiled-debug-info-source obj)
-                                   canonical-repr)))))))
-               (loop for debug-fun = (compiled-debug-info-fun-map obj) then next
-                     for next = (sb-c::compiled-debug-fun-next debug-fun)
-                     do
-                     (binding* ((name (compiled-debug-fun-name debug-fun))
-                                ((new foundp) (gethash name name-ht)))
-                       (cond ((not foundp)
-                              (setf (gethash name name-ht) name))
-                             ((neq name new)
-                              (%instance-set debug-fun (get-dsd-index compiled-debug-fun name)
-                                             new))))
-                     while next))
+                                   canonical-repr))))))))
               (sb-lockless::linked-list
                ;; In the normal course of execution, incompletely deleted nodes
                ;; exist only for a brief moment, as the next operation on the list by
@@ -531,7 +517,7 @@ sufficiently motivated to do lengthy fixes."
                           (and (typep di 'sb-c::compiled-debug-info)
                                (let ((src (sb-c::compiled-debug-info-source di)))
                                  (and (typep src 'sb-c::debug-source)
-                                      (let ((str (debug-source-namestring src)))
+                                      (let ((str (sb-c::debug-source-namestring src)))
                                         (if (= (mismatch str "SYS:") 4) 1))))))
                         0))
                    ;; cap the popularity index to 255 and negate so that higher
