@@ -248,38 +248,20 @@
      (setf (%simple-fun-arglist (%fun-fun function)) new-value)))
   new-value)
 
-(macrolet ((access-slot (index)
-             `(code-header-ref
-               (fun-code-header fun)
-               (+ (* sb-vm:code-slots-per-simple-fun (%simple-fun-index fun))
-                  sb-vm:code-constants-offset ,index)))
-           (def (accessor index)
+(macrolet ((index (slot)
+             `(+ (* sb-vm:code-slots-per-simple-fun (%simple-fun-index fun))
+                 (get-dsd-index sb-c::compiled-debug-info rest)
+                 ,slot))
+           (def (accessor slot)
              `(progn
                 (defun (setf ,accessor) (newval fun)
                   (declare (simple-fun fun))
-                  ;; Prevent wild pointers due to 'purify' moving all code to
-                  ;; readonly space. (Can't have read-only pointing to dynamic)
-                  ;; There are a number of things we could do to "fix" this, none
-                  ;; particularly interesting or meritorious imho, e.g.:
-                  ;;   - Copy the name to static space
-                  ;;   - Use an external hash-table (a la named closures),
-                  ;;   - Implement some other notion of "forwarded" names
-                  ;;   - Track which pages of read-only [sic] space have been written
-                  ;;   - Scavenge all of read-only space always
-                  ;;   - Write-protect read-only space to completely prevent this
-                  (if #+cheneygc (and (eq (heap-allocated-p fun) :read-only)
-                                      (eq (heap-allocated-p newval) :dynamic))
-                      #-cheneygc nil
-                      (progn (warn ,(format nil "Can't assign ~A of ~~A" accessor) fun)
-                             newval)
-                      (setf (access-slot ,index) newval)))
+                  (setf (%instance-ref (%code-debug-info (fun-code-header fun))
+                                       (index ,slot))
+                        newval))
                 (defun ,accessor (fun)
-                  (access-slot ,index)))))
-  ;; possible FIXME for the backends which treat the assembly trampolines
-  ;; as tagged functions (with fun-pointer-lowtag) - we might need to ensure
-  ;; that the code object reserves space for 4 NILs just in case a simple-fun
-  ;; accessor is called on it. I'm not entirely sure whether that's necessary.
-
+                  (%instance-ref (%code-debug-info (fun-code-header fun))
+                                 (index ,slot))))))
   (def %simple-fun-name    sb-vm:simple-fun-name-slot)
   (def %simple-fun-arglist sb-vm:simple-fun-arglist-slot)
   (def %simple-fun-source  sb-vm:simple-fun-source-slot)
@@ -328,7 +310,7 @@
                (context
                 (sb-kernel::make-type-context
                  ftype nil sb-kernel::+type-parse-signal-inhibit+)))
-          (declare (truly-dynamic-extent context))
+          (declare (dynamic-extent context))
           (values (sb-kernel::basic-parse-typespec ftype context)))))))
 
 ;;; Return the lambda expression for SIMPLE-FUN if compiled to memory
@@ -383,7 +365,7 @@
   newval)
 
 (defun (setf sb-vm::%code-fixups) (newval code)
-  (code-header-set code sb-vm::code-fixups-slot newval)
+  (setf (code-header-ref code sb-vm::code-fixups-slot) newval)
   newval)
 
 #+(or sparc ppc64)
@@ -424,8 +406,7 @@
 ;;; (i.e. at the head of an expression)
 (defun code-header-fdefn-range (code-obj)
   #-64-bit ; inefficient
-  (let ((start (+ sb-vm:code-constants-offset
-                  (* (code-n-entries code-obj) sb-vm:code-slots-per-simple-fun)))
+  (let ((start sb-vm:code-constants-offset)
         (count 0))
     (do ((i start (1+ i))
          (limit (code-header-words code-obj)))
@@ -433,8 +414,7 @@
       (if (fdefn-p (code-header-ref code-obj i)) (incf count) (return)))
     (values start count))
   #+64-bit
-  (values (+ sb-vm:code-constants-offset
-             (* (code-n-entries code-obj) sb-vm:code-slots-per-simple-fun))
+  (values sb-vm:code-constants-offset
           (ash (sb-vm::%code-boxed-size code-obj)
                (+ -32 sb-vm:n-fixnum-tag-bits))))
 
@@ -579,21 +559,8 @@
              (error (if (eq kind :special) 'special-form-function 'undefined-function)
                     :name  (second fun-name))))
           t
-          fun-name))
-        (fdefn (find-or-create-fdefn symbol)))
-
-    ;; In most cases, install the guard closure in the usual way.
-    #-(and x86-64 immobile-code) (setf (fdefn-fun fdefn) closure)
-
-    ;; Do something slightly different for immobile code: fmakunbound, assigning
-    ;; FUN = NIL and RAW-ADDR = UNDEFINED-TRAMP; then overwrite the NIL with the
-    ;; above closure. This is better than assigning a closure, because closures
-    ;; require a new closure-calling trampoline to be consed.
-    #+(and x86-64 immobile-code)
-    (progn (fdefn-makunbound fdefn)
-           (%primitive sb-vm::set-undefined-fdefn-fun fdefn closure))
-
-    fdefn))
+          fun-name)))
+    (fset symbol closure)))
 
 ;;;; Iterating over closure values
 
