@@ -320,9 +320,10 @@ corresponds to NAME, or NIL if there is none."
 
 (defun unix-read (fd buf len)
   (declare (type unix-fd fd)
-           (type (unsigned-byte 32) len))
+           (type index len))
   (int-syscall (#-win32 "read" #+win32 "win32_unix_read"
-                int (* char) int) fd buf len))
+                ssize-t (* char) size-t)
+               fd buf len))
 
 ;;; UNIX-WRITE accepts a file descriptor, a buffer, an offset, and the
 ;;; length to write. It attempts to write len bytes to the device
@@ -334,11 +335,11 @@ corresponds to NAME, or NIL if there is none."
   ;; full calls to SB-ALIEN-INTERNALS:DEPORT-ALLOC and DEPORT.
   (declare (optimize (debug 1)))
   (declare (type unix-fd fd)
-           (type (unsigned-byte 32) offset len))
+           (type index offset len))
   (flet ((%write (sap)
            (declare (system-area-pointer sap))
            (int-syscall (#-win32 "write" #+win32 "win32_unix_write"
-                         int (* char) int)
+                         ssize-t (* char) size-t)
                         fd
                         (with-alien ((ptr (* char) sap))
                           (addr (deref ptr offset)))
@@ -551,7 +552,7 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 (defun unix-ioctl (fd cmd arg)
   (declare (type unix-fd fd)
            (type word cmd))
-  (void-syscall ("ioctl" int unsigned-long (* char)) fd cmd arg))
+  (void-syscall ("ioctl" int unsigned-long &optional (* char)) fd cmd arg))
 
 ;;;; sys/resource.h
 
@@ -870,8 +871,9 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 
 (define-alien-routine get-timezone int
   (when time-t)
-  ;; KLUDGE: the runtime `boolean' is defined as `int', but the alien
-  ;; type is N-WORD-BITS wide.
+  ;; BOOLEAN is N-WORD-BITS normally. Reduce it to an unsigned int in size.
+  ;; But we can't just put UNSIGNED-INT here because the clients of this function
+  ;; want to receive a T or NIL, not a 1 or 0.
   (daylight-savings-p (boolean 32) :out))
 #-win32
 (defun nanosleep (secs nsecs)
@@ -986,7 +988,9 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 #-win32
 (progn
 
+  #-avoid-clock-gettime
   (declaim (inline clock-gettime))
+  #-avoid-clock-gettime
   (defun clock-gettime (clockid)
     (declare (type (signed-byte 32) clockid))
     (with-alien ((ts (struct timespec)))
@@ -1021,7 +1025,10 @@ the UNIX epoch (January 1st 1970.)"
           ;; By scaling down we end up with far less resolution than clock-realtime
           ;; offers, and COARSE is about twice as fast, so use that, but only for linux.
           ;; BSD has something similar.
+          #-avoid-clock-gettime
           (clock-gettime #+linux clock-monotonic-coarse #-linux clock-monotonic)
+          #+avoid-clock-gettime
+          (multiple-value-bind (c-sec c-usec) (get-time-of-day) (values c-sec (* c-usec 1000)))
 
         #+64-bit ;; I know that my math is valid for 64-bit.
         (declare (optimize (sb-c::type-check 0)))
@@ -1070,12 +1077,29 @@ the UNIX epoch (January 1st 1970.)"
                   current)))))))
 
   (declaim (inline system-internal-run-time))
-  #-sunos ; defined in sunos-os
+
+  ;; SunOS defines CLOCK_PROCESS_CPUTIME_ID but you get EINVAL if you try to use it,
+  ;; also use the same trick when clock_gettime should be avoided.
+  #-(or sunos avoid-clock-gettime)
   (defun system-internal-run-time ()
     (multiple-value-bind (sec nsec) (clock-gettime clock-process-cputime-id)
       (+ (* sec internal-time-units-per-second)
          (floor (+ nsec (floor nanoseconds-per-internal-time-unit 2))
-                nanoseconds-per-internal-time-unit)))))
+                nanoseconds-per-internal-time-unit))))
+  #+(or sunos avoid-clock-gettime)
+  (defun system-internal-run-time ()
+    (multiple-value-bind (utime-sec utime-usec stime-sec stime-usec)
+        (with-alien ((usage (struct sb-unix::rusage)))
+          (syscall* ("sb_getrusage" int (* (struct sb-unix::rusage)))
+                    (values (slot (slot usage 'sb-unix::ru-utime) 'sb-unix::tv-sec)
+                            (slot (slot usage 'sb-unix::ru-utime) 'sb-unix::tv-usec)
+                            (slot (slot usage 'sb-unix::ru-stime) 'sb-unix::tv-sec)
+                            (slot (slot usage 'sb-unix::ru-stime) 'sb-unix::tv-usec))
+                    rusage_self (addr usage)))
+      (+ (* (+ utime-sec stime-sec) internal-time-units-per-second)
+         (floor (+ utime-usec stime-usec
+                   (floor microseconds-per-internal-time-unit 2))
+                microseconds-per-internal-time-unit)))))
 
 ;;; FIXME, KLUDGE: GET-TIME-OF-DAY used to be UNIX-GETTIMEOFDAY, and had a
 ;;; primary return value indicating sucess, and also returned timezone
