@@ -108,8 +108,8 @@
 
 (defun optional-leaf-p (leaf)
   (let ((home (lambda-var-home leaf)))
-    (case (functional-kind home)
-      (:external
+    (functional-kind-case home
+      (external
        (let ((entry (lambda-entry-fun home)))
          (when (optional-dispatch-p entry)
            (let ((pos (position leaf (lambda-vars home))))
@@ -122,8 +122,8 @@
 ;;; optional is alive at that point.
 (defun optional-processed (leaf)
   (let ((home (lambda-var-home leaf)))
-    (case (functional-kind home)
-      (:external
+    (functional-kind-case home
+      (external
        (let ((entry (lambda-entry-fun home)))
          (if (and (optional-dispatch-p entry)
                   *local-call-context*)
@@ -185,6 +185,13 @@
            (the fixnum (logior (ash offset registers-size)
                                (tn-offset tn)))
            offset)))
+    (cons (let ((last (cdr (last x))))
+            (if (restart-location-p last)
+                (let ((new (copy-list x)))
+                  (setf (cdr (last new))
+                        (encode-restart-location location last))
+                  new)
+                x)))
     (t
      x)))
 
@@ -366,9 +373,7 @@
     (when elsewhere-locations
       (dolist (loc (nreverse elsewhere-locations))
         (dump-location-from-info loc tlf-num var-locs)))
-    ;; lz-compress accept any array of octets and returns a simple-array
-    (values (logically-readonlyize (lz-compress byte-buffer))
-            tlf-num)))
+    (values (coerce byte-buffer '(simple-array (unsigned-byte 8) (*))) tlf-num)))
 
 ;;; Return DEBUG-SOURCE structure containing information derived from
 ;;; INFO.
@@ -385,7 +390,19 @@
                                (truename (file-info-truename file-info)))))
                         (if (pathnamep pathname) pathname))
                       file-info))
-      :created (file-info-write-date file-info)
+      :created
+      #+sb-xc-host (file-info-write-date file-info)
+      #-sb-xc-host (let ((source-date-epoch (posix-getenv "SOURCE_DATE_EPOCH"))
+                         (file-write-date (file-info-write-date file-info)))
+                     (if source-date-epoch
+                         (multiple-value-bind (val end)
+                             (parse-integer source-date-epoch :junk-allowed t)
+                           (if (and (= end (length source-date-epoch))
+                                    (and val file-write-date)
+                                    (< (+ val unix-to-universal-time) file-write-date))
+                               (+ val unix-to-universal-time)
+                               file-write-date))
+                         file-write-date))
       :start-positions (coerce-to-smallest-eltype
                         (file-info-positions file-info))
      (if function
@@ -413,6 +430,7 @@
 ;;; During cross-compilation the in-memory representation is opaque -
 ;;; we don't care how it looks, but can recover the intended specialization.
 (defun coerce-to-smallest-eltype (seq)
+  (declare (type sequence seq))
   (let ((max-positive 0)
         (max-negative 0)
         (length 0))
@@ -472,22 +490,31 @@
 (defun dump-1-var (fun var tn minimal buffer &optional name same-name-p)
   (declare (type lambda-var var) (type (or tn null) tn)
            (type clambda fun))
-  (let* ((save-tn (and tn (tn-save-tn tn)))
+  (let* ((package (sb-xc:symbol-package name))
+         (package-p (and package (not (eq package *package*))))
+         (save-tn (and tn (tn-save-tn tn)))
          (kind (and tn (tn-kind tn)))
          (flags 0)
-         (info (lambda-var-arg-info var))
          (indirect (and (lambda-var-indirect var)
                         (not (lambda-var-explicit-value-cell var))
                         (neq (lambda-environment fun)
-                             (lambda-environment (lambda-var-home var)))))
-         ;; Keep this condition in sync with PARSE-COMPILED-DEBUG-VARS
-         (large-fixnums (>= (integer-length most-positive-fixnum) 62))
-         more)
+                             (lambda-environment (lambda-var-home var))))))
     (declare (type (and sb-xc:fixnum unsigned-byte) flags))
-    (when minimal
-      (setq flags (logior flags compiled-debug-var-minimal-p))
-      (unless (and tn (tn-offset tn))
-        (setq flags (logior flags compiled-debug-var-deleted-p))))
+    (let ((info (lambda-var-arg-info var)))
+      ;; &more vars need no name
+      (when (and info
+                 (memq (arg-info-kind info)
+                       '(:more-context :more-count)))
+        (setq minimal t)))
+    (cond (minimal
+           (setq flags (logior flags compiled-debug-var-minimal-p))
+           (unless (and tn (tn-offset tn))
+             (setq flags (logior flags compiled-debug-var-deleted-p))))
+          (t
+           (unless package
+             (setq flags (logior flags compiled-debug-var-uninterned)))
+           (when package-p
+             (setq flags (logior flags compiled-debug-var-packaged)))))
     (when (and (or (eq kind :environment)
                    (and (eq kind :debug-environment)
                         (null (basic-var-sets var))))
@@ -501,56 +528,30 @@
       (setq flags (logior flags compiled-debug-var-save-loc-p)))
     (when indirect
       (setq flags (logior flags compiled-debug-var-indirect-p)))
-    (when info
-      (case (arg-info-kind info)
-        (:more-context
-         (setq flags (logior flags compiled-debug-var-more-context-p)
-               more t))
-        (:more-count
-         (setq flags (logior flags compiled-debug-var-more-count-p)
-               more t))))
-    (when (and same-name-p
-               (not (or more minimal)))
-      (setf flags (logior flags compiled-debug-var-same-name-p)))
-    (when large-fixnums
-      (cond (indirect
-             (setf (ldb (byte 27 8) flags) (tn-sc+offset tn))
-             (when save-tn
-               (setf (ldb (byte 27 35) flags) (tn-sc+offset save-tn))))
-            (t
-             (if (and tn (tn-offset tn))
-                 (setf (ldb (byte 27 8) flags) (tn-sc+offset tn))
-                 (aver minimal))
-             (when save-tn
-               (setf (ldb (byte 27 35) flags) (tn-sc+offset save-tn))))))
+    (when (and same-name-p (not minimal))
+      (setq flags (logior flags compiled-debug-var-same-name-p)))
     (vector-push-extend flags buffer)
-    (unless (or minimal
-                same-name-p
-                more) ;; &more vars need no name
-      ;; Dumping uninterned symbols as debug var names is kinda silly.
-      ;; Reconstruction of the name on fasl load produces a new gensym anyway.
-      ;; So rather than waste symbol space, just dump such symbols as strings,
-      ;; and PARSE-COMPILED-DEBUG-VARS can create the interned symbol.
-      ;; This reduces core size by omitting zillions of symbols whose names
-      ;; are spelled the same.
-      (vector-push-extend (if (cl:symbol-package name) name (string name)) buffer))
+    (unless minimal
+      (unless same-name-p
+        (write-var-string (symbol-name name) buffer))
+      (when package-p
+        (write-var-string (sb-xc:package-name package) buffer)))
 
     (cond (indirect
            ;; Indirect variables live in the parent frame, and are
            ;; accessed through a saved frame pointer.
            ;; The first one/two sc-offsets are for the frame pointer,
            ;; the third is for the stack offset.
-           (unless large-fixnums
-             (vector-push-extend (tn-sc+offset tn) buffer)
-             (when save-tn
-               (vector-push-extend (tn-sc+offset save-tn) buffer)))
-           (vector-push-extend (tn-sc+offset (leaf-info var)) buffer))
-          ((not large-fixnums)
+           (write-var-integer (tn-sc+offset tn) buffer)
+           (when save-tn
+             (write-var-integer (tn-sc+offset save-tn) buffer))
+           (write-var-integer (tn-sc+offset (leaf-info var)) buffer))
+          (t
            (if (and tn (tn-offset tn))
-               (vector-push-extend (tn-sc+offset tn) buffer)
+               (write-var-integer (tn-sc+offset tn) buffer)
                (aver minimal))
            (when save-tn
-             (vector-push-extend (tn-sc+offset save-tn) buffer)))))
+             (write-var-integer (tn-sc+offset save-tn) buffer)))))
   (values))
 
 (defun leaf-principal-name (leaf)
@@ -591,36 +592,36 @@
         (dolist (let (lambda-lets fun))
           (frob-lambda let (>= level 2)))))
 
+    (setf (fill-pointer *byte-buffer*) 0)
     (let ((sorted (sort (vars) #'string<
                         :key (lambda (x)
                                (symbol-name (car x)))))
           (prev-name nil)
           (i 0)
-          (buffer (make-array 0 :fill-pointer 0 :adjustable t))
           ;; XEPs don't have any useful variables
-          (minimal (eq (functional-kind fun) :external)))
+          (minimal (functional-kind-eq fun external)))
       (declare (type index i))
       (loop for (name var . tn) in sorted
             do
-            (dump-1-var fun var tn minimal buffer
+            (dump-1-var fun var tn minimal *byte-buffer*
                         name
                         (and prev-name (eq prev-name name)))
             (setf prev-name name)
             (setf (gethash var var-locs) i)
             (incf i))
-      (compact-vector buffer))))
+      (copy-seq *byte-buffer*))))
 
 ;;; Return a vector suitable for use as the DEBUG-FUN-VARS of
 ;;; FUN, representing the arguments to FUN in minimal variable format.
 (defun compute-minimal-vars (fun)
   (declare (type clambda fun))
-  (let ((buffer (make-array 0 :fill-pointer 0 :adjustable t)))
-    (dolist (var (lambda-vars fun))
-      (dump-1-var fun var (leaf-info var) t buffer))
-    (compact-vector buffer)))
+  (setf (fill-pointer *byte-buffer*) 0)
+  (dolist (var (lambda-vars fun))
+    (dump-1-var fun var (leaf-info var) t *byte-buffer*))
+  (copy-seq *byte-buffer*))
 
 ;;; Return VAR's relative position in the function's variables (determined
-;;; from the VAR-LOCS hashtable).  If VAR is deleted, then return DEBUG-INFO-VAR-DELETED.
+;;; from the VAR-LOCS hashtable).  If VAR is deleted, then return DELETED.
 (defun debug-location-for (var var-locs)
   (declare (type lambda-var var) (type hash-table var-locs))
   (let ((res (gethash var var-locs)))
@@ -628,7 +629,7 @@
           (t
            (aver (or (null (leaf-refs var))
                      (not (tn-offset (leaf-info var)))))
-           debug-info-var-deleted))))
+           'deleted))))
 
 ;;;; arguments/returns
 
@@ -661,16 +662,16 @@
                                             (return-from one-arg))
                                            (more
                                             (setf (arg-info-default info) t)))
-                                     (res debug-info-var-rest)))
+                                     (res 'rest-arg)))
                                   (:more-context
-                                   (res debug-info-var-more))
+                                   (res 'more-arg))
                                   (:optional
                                    (unless saw-optional
-                                     (res debug-info-var-optional)
+                                     (res 'optional-args)
                                      (setq saw-optional t))))
                                 (res (debug-location-for actual var-locs))
                                 (when (arg-info-supplied-p info)
-                                  (res debug-info-var-supplied-p)
+                                  (res 'supplied-p)
                                   (res (debug-location-for (pop actual-vars) var-locs))))
                                 (t
                                  (res (debug-location-for actual var-locs)))))))
@@ -679,7 +680,7 @@
           (dolist (var (lambda-vars fun))
             (res (debug-location-for var var-locs)))))
 
-    (compact-vector (res))))
+    (coerce-to-smallest-eltype (res))))
 
 ;;; Return a vector of SC offsets describing FUN's return locations.
 ;;; (Must be known values return...)
@@ -697,12 +698,17 @@
          (dispatch (lambda-optional-dispatch fun))
          (main-p (and dispatch
                       (eq fun (optional-dispatch-main-entry dispatch))))
-         (kind (if main-p nil (functional-kind fun)))
+         (kind (if main-p nil (ecase (functional-kind fun)
+                                (#.(functional-kind-attributes nil) nil)
+                                (#.(functional-kind-attributes optional) :optional)
+                                (#.(functional-kind-attributes external) :external)
+                                (#.(functional-kind-attributes toplevel) :toplevel)
+                                (#.(functional-kind-attributes cleanup) :cleanup))))
          (name (leaf-debug-name fun))
          (name (if (consp name)
                    (case (car name)
                      ((xep tl-xep)
-                      (aver (eq kind :external))
+                      (aver (eql kind :external))
                       (second name))
                      (&optional-processor
                       (setf kind :optional)
@@ -713,29 +719,30 @@
                      (t
                       name))
                    name)))
-    (funcall (compiled-debug-fun-ctor kind)
-             :name name
-             #-fp-and-pc-standard-save :return-pc
-             #-fp-and-pc-standard-save (tn-sc+offset (ir2-environment-return-pc 2env))
-             #-fp-and-pc-standard-save :return-pc-pass
-             #-fp-and-pc-standard-save (tn-sc+offset (ir2-environment-return-pc-pass 2env))
-             #-fp-and-pc-standard-save :old-fp
-             #-fp-and-pc-standard-save (tn-sc+offset (ir2-environment-old-fp 2env))
-             :encoded-locs
-             (cdf-encode-locs
-              (label-position (ir2-environment-environment-start 2env))
-              (label-position (ir2-environment-elsewhere-start 2env))
-              (source-path-form-number (node-source-path (lambda-bind fun)))
-              (label-position (block-label (lambda-block fun)))
-              (when (ir2-environment-closure-save-tn 2env)
-                (tn-sc+offset (ir2-environment-closure-save-tn 2env)))
-              #+unwind-to-frame-and-call-vop
-              (when (ir2-environment-bsp-save-tn 2env)
-                (tn-sc+offset (ir2-environment-bsp-save-tn 2env)))
-              #-fp-and-pc-standard-save
-              (label-position (ir2-environment-lra-saved-pc 2env))
-              #-fp-and-pc-standard-save
-              (label-position (ir2-environment-cfp-saved-pc 2env))))))
+    (make-compiled-debug-fun
+     :name name
+     :kind kind
+     #-fp-and-pc-standard-save :return-pc
+     #-fp-and-pc-standard-save (tn-sc+offset (ir2-environment-return-pc 2env))
+     #-fp-and-pc-standard-save :return-pc-pass
+     #-fp-and-pc-standard-save (tn-sc+offset (ir2-environment-return-pc-pass 2env))
+     #-fp-and-pc-standard-save :old-fp
+     #-fp-and-pc-standard-save (tn-sc+offset (ir2-environment-old-fp 2env))
+     #-fp-and-pc-standard-save :lra-saved-pc
+     #-fp-and-pc-standard-save (label-position (ir2-environment-lra-saved-pc 2env))
+     #-fp-and-pc-standard-save :cfp-saved-pc
+     #-fp-and-pc-standard-save (label-position (ir2-environment-cfp-saved-pc 2env))
+     :closure-save
+     (when (ir2-environment-closure-save-tn 2env)
+       (tn-sc+offset (ir2-environment-closure-save-tn 2env)))
+     #+unwind-to-frame-and-call-vop :bsp-save
+     #+unwind-to-frame-and-call-vop
+     (when (ir2-environment-bsp-save-tn 2env)
+       (tn-sc+offset (ir2-environment-bsp-save-tn 2env)))
+     :start-pc
+     (label-position (ir2-environment-environment-start 2env))
+     :elsewhere-pc
+     (label-position (ir2-environment-elsewhere-start 2env)))))
 
 ;;; Return a complete C-D-F structure for FUN. This involves
 ;;; determining the DEBUG-INFO level and filling in optional slots as
@@ -768,6 +775,7 @@
           (setf (compiled-debug-fun-blocks dfun) blocks
                 (compiled-debug-fun-tlf-number dfun) tlf-num))
         (setf (compiled-debug-fun-tlf-number dfun) (find-tlf-number fun)))
+
     (if (xep-p fun)
         (setf (compiled-debug-fun-returns dfun) :standard)
         (let ((info (tail-set-info (lambda-tail-set fun))))
@@ -780,16 +788,170 @@
                    (setf (compiled-debug-fun-returns dfun)
                          (compute-debug-returns fun)))))))
     dfun))
+
+
+;;;; Packed debug functions:
+
+;;; Dump a packed binary representation of a DFUN into *BYTE-BUFFER*.
+;;; PREV-START and START are the byte offsets in the code where the
+;;; previous function started and where this one starts.
+;;; PREV-ELSEWHERE is the previous function's elsewhere PC.
+(defun dump-1-packed-dfun (dfun prev-start start prev-elsewhere)
+  (declare (type compiled-debug-fun dfun)
+           (type index prev-start start prev-elsewhere)
+           (special *debug-component-name*))
+  (let* ((name (compiled-debug-fun-name dfun))
+         (same-name-p (eq name
+                          *debug-component-name*)))
+    (let ((options 0))
+      (setf (ldb packed-debug-fun-kind-byte options)
+            (position-or-lose (compiled-debug-fun-kind dfun)
+                              packed-debug-fun-kinds))
+      (setf (ldb packed-debug-fun-returns-byte options)
+            (etypecase (compiled-debug-fun-returns dfun)
+              ((member :standard) packed-debug-fun-returns-standard)
+              ((member :fixed) packed-debug-fun-returns-fixed)
+              (vector packed-debug-fun-returns-specified)))
+      (vector-push-extend options *byte-buffer*))
+
+    (let ((flags 0))
+      (when (compiled-debug-fun-vars dfun)
+        (setq flags (logior flags packed-debug-fun-variables-bit)))
+      (when (compiled-debug-fun-blocks dfun)
+        (setq flags (logior flags packed-debug-fun-blocks-bit)))
+      (when (compiled-debug-fun-tlf-number dfun)
+        (setq flags (logior flags packed-debug-fun-tlf-number-bit)))
+      (unless (memq (compiled-debug-fun-arguments dfun) '(nil :minimal))
+        (setq flags (logior flags packed-debug-fun-non-minimal-arguments-bit)))
+      (when (compiled-debug-fun-closure-save dfun)
+        (setq flags (logior flags packed-debug-fun-closure-save-loc-bit)))
+      #+unwind-to-frame-and-call-vop
+      (when (compiled-debug-fun-bsp-save dfun)
+        (setq flags (logior flags packed-debug-fun-bsp-save-loc-bit)))
+      (when same-name-p
+        (setq flags (logior flags packed-debug-fun-previous-name)))
+      (vector-push-extend flags *byte-buffer*))
+    (unless same-name-p
+      (setf *debug-component-name* name)
+      (write-var-integer (or (position name *contexts* :test #'equal)
+                             (vector-push-extend name *contexts*))
+                         *byte-buffer*))
+
+    (let ((vars (compiled-debug-fun-vars dfun)))
+      (when vars
+        (let ((len (length vars)))
+          (write-var-integer len *byte-buffer*)
+          (dotimes (i len)
+            (vector-push-extend (aref vars i) *byte-buffer*)))))
+
+    (let ((blocks (compiled-debug-fun-blocks dfun)))
+      (when blocks
+        (let ((len (length blocks)))
+          (write-var-integer len *byte-buffer*)
+          (dotimes (i len)
+            (vector-push-extend (mask-field (byte 8 0) (aref blocks i))
+                                *byte-buffer*)))))
+
+    (when (compiled-debug-fun-tlf-number dfun)
+      (write-var-integer (compiled-debug-fun-tlf-number dfun)
+                         *byte-buffer*))
+
+    (let ((arguments (compiled-debug-fun-arguments dfun)))
+      (unless (memq arguments '(nil :minimal))
+        (let ((len (length arguments)))
+          (write-var-integer len *byte-buffer*)
+          (dotimes (i len)
+            (let ((argument (aref arguments i)))
+              (case argument
+                (deleted
+                 (write-var-integer packed-debug-fun-arg-deleted
+                                    *byte-buffer*))
+                (supplied-p
+                 (write-var-integer packed-debug-fun-arg-supplied-p
+                                    *byte-buffer*))
+                (optional
+                 (write-var-integer packed-debug-fun-arg-optional
+                                    *byte-buffer*))
+                (rest
+                 (write-var-integer packed-debug-fun-arg-rest
+                                    *byte-buffer*))
+                (more
+                 (write-var-integer packed-debug-fun-arg-more
+                                    *byte-buffer*))
+                (otherwise
+                 (cond ((integerp argument)
+                        (write-var-integer (+ argument
+                                              packed-debug-fun-arg-index-offset)
+                                           *byte-buffer*))
+                       ((keywordp argument)
+                        (write-var-integer packed-debug-fun-key-arg-keyword
+                                           *byte-buffer*)
+                        (write-var-string (symbol-name argument)
+                                          *byte-buffer*))
+                       ((sb-xc:symbol-package argument)
+                        (write-var-integer packed-debug-fun-key-arg-packaged
+                                           *byte-buffer*)
+                        (write-var-string (symbol-name argument)
+                                          *byte-buffer*)
+                        (write-var-string (sb-xc:package-name
+                                           (sb-xc:symbol-package argument))
+                                          *byte-buffer*))
+                       (t
+                        (write-var-integer packed-debug-fun-key-arg-uninterned
+                                           *byte-buffer*)
+                        (write-var-string (symbol-name argument)
+                                          *byte-buffer*))))))))))
+
+    (let ((returns (compiled-debug-fun-returns dfun)))
+      (when (vectorp returns)
+        (let ((len (length returns)))
+          (write-var-integer len *byte-buffer*)
+          (dotimes (i len)
+            (write-var-integer (aref returns i) *byte-buffer*)))))
+
+    #-fp-and-pc-standard-save
+    (progn
+      (write-var-integer (compiled-debug-fun-return-pc dfun)
+                         *byte-buffer*)
+      (write-var-integer (compiled-debug-fun-return-pc-pass dfun)
+                         *byte-buffer*)
+      (write-var-integer (compiled-debug-fun-old-fp dfun)
+                         *byte-buffer*)
+      (write-var-integer (compiled-debug-fun-lra-saved-pc dfun)
+                         *byte-buffer*)
+      (write-var-integer (compiled-debug-fun-cfp-saved-pc dfun)
+                         *byte-buffer*))
+    (when (compiled-debug-fun-closure-save dfun)
+      (write-var-integer (compiled-debug-fun-closure-save dfun)
+                         *byte-buffer*))
+    #+unwind-to-frame-and-call-vop
+    (when (compiled-debug-fun-bsp-save dfun)
+      (write-var-integer (compiled-debug-fun-bsp-save dfun)
+                         *byte-buffer*))
+    (write-var-integer (- start prev-start) *byte-buffer*)
+    (write-var-integer (- (compiled-debug-fun-start-pc dfun) start)
+                       *byte-buffer*)
+    (write-var-integer (- (compiled-debug-fun-elsewhere-pc dfun)
+                          prev-elsewhere)
+                       *byte-buffer*)))
+
+;;; Return a byte-vector holding all the debug functions for a
+;;; component in the packed binary PACKED-DEBUG-FUN format.
+(defun compute-packed-debug-funs (dfuns)
+  (declare (list dfuns))
+  (setf (fill-pointer *byte-buffer*) 0)
+  (let ((prev-start 0)
+        (prev-elsewhere 0))
+    (dolist (dfun dfuns)
+      (let ((start (car dfun))
+            (elsewhere (compiled-debug-fun-elsewhere-pc (cdr dfun))))
+        (dump-1-packed-dfun (cdr dfun) prev-start start prev-elsewhere)
+        (setq prev-start start prev-elsewhere elsewhere))))
+  (logically-readonlyize (compress *byte-buffer*)))
+
 
 ;;;; full component dumping
 
-;;; Compute the full form function map.
-(defun compute-debug-fun-map (sorted)
-  (declare (list sorted))
-  (loop for (fun next) on sorted
-        do (setf (compiled-debug-fun-next fun) next))
-  (car sorted))
-
 (defun empty-fun-p (fun)
   (let* ((2block (block-info (lambda-block fun)))
          (start (ir2-block-start-vop 2block))
@@ -806,34 +968,63 @@
 ;;; called after assembly so that source map information is available.
 (defun debug-info-for-component (component)
   (declare (type component component))
-  (let ((dfuns nil)
-        (var-locs (make-hash-table :test 'eq))
-        (*byte-buffer* (make-array 10
-                                   :element-type '(unsigned-byte 8)
-                                   :fill-pointer 0
-                                   :adjustable t))
-        (*contexts* (make-array 10
-                                :fill-pointer 0
-                                :adjustable t)))
-    (dolist (lambda (component-lambdas component))
+  (let* ((dfuns nil)
+         (simple-fun-headers
+          ;; Compute all simple-fun metadata and store into a simple-vector
+          (let* ((entries (ir2-component-entries (component-info component)))
+                 (nfuns (length entries))
+                 (i (* sb-vm:code-slots-per-simple-fun nfuns))
+                 (v (make-array i)))
+            (dolist (e entries v)
+              ;; Process in reverse order of ENTRIES.
+              (decf i sb-vm:code-slots-per-simple-fun)
+              (setf (svref v (+ i sb-vm:simple-fun-name-slot)) (entry-info-name e)
+                    (svref v (+ i sb-vm:simple-fun-arglist-slot)) (entry-info-arguments e)
+                    (svref v (+ i sb-vm:simple-fun-source-slot)) (entry-info-form/doc e)
+                    (svref v (+ i sb-vm:simple-fun-info-slot)) (entry-info-type/xref e)))))
+         (var-locs (make-hash-table :test 'eq))
+         (*byte-buffer* (make-array 10
+                                    :element-type '(unsigned-byte 8)
+                                    :fill-pointer 0
+                                    :adjustable t))
+         (*contexts* (make-array 10
+                                 :fill-pointer 0
+                                 :adjustable t))
+         (lambdas (sort (copy-list (component-lambdas component))
+                        #'<
+                        :key (lambda (lambda)
+                               (label-position (block-label (lambda-block lambda))))))
+         (name (loop for lambda in lambdas
+                     for entry = (leaf-info lambda)
+                     when entry
+                     return
+                     (entry-info-name entry)))
+         (*debug-component-name* name))
+    (declare (special *debug-component-name*))
+    (dolist (lambda lambdas)
       (unless (empty-fun-p lambda)
         (clrhash var-locs)
-        (push (compute-1-debug-fun lambda var-locs)
+        (push (cons (label-position (block-label (lambda-block lambda)))
+                    (compute-1-debug-fun lambda var-locs))
               dfuns)))
-    (let* ((sorted (sort dfuns #'< :key #'compiled-debug-fun-offset))
-           (fun-map (compute-debug-fun-map sorted)))
-      (make-compiled-debug-info
-       ;; COMPONENT-NAME is often not useful, and sometimes completely fubar.
-       ;; Function names, on the other hand, are seldom unhelpful,
-       ;; so if there's only one function, pick that as the component name.
-       ;; Otherwise preserve whatever crummy name was already assigned.
-       :name (let* ((2comp (component-info component))
-                    (entries (sb-c::ir2-component-entries 2comp)))
-               (or (and (not (cdr entries))
-                        (sb-c::entry-info-name (car entries)))
-                   (component-name component)))
-       :fun-map fun-map
-       :contexts (compact-vector *contexts*)))))
+    (let ((map (compute-packed-debug-funs (nreverse dfuns)))
+          (contexts (compact-vector *contexts*)))
+      #+sb-xc-host
+      (!make-compiled-debug-info name *package* map contexts simple-fun-headers)
+      #-sb-xc-host
+      (let ((di (%make-instance (+ (sb-kernel::type-dd-length compiled-debug-info)
+                                   (length simple-fun-headers)))))
+        (setf (%instance-layout di) #.(find-layout 'compiled-debug-info)
+              ;; The fixed slots except for SOURCE are declared readonly
+              (%instance-ref di (get-dsd-index compiled-debug-info name)) name
+              (%instance-ref di (get-dsd-index compiled-debug-info source)) nil
+              (%instance-ref di (get-dsd-index compiled-debug-info package)) *package*
+              (%instance-ref di (get-dsd-index compiled-debug-info fun-map)) map
+              (%instance-ref di (get-dsd-index compiled-debug-info contexts)) contexts)
+        (let ((i (get-dsd-index compiled-debug-info rest)))
+          (dovector (x simple-fun-headers di)
+            (setf (%instance-ref di i) x)
+            (incf i)))))))
 
 ;;; Write BITS out to BYTE-BUFFER in backend byte order. The length of
 ;;; BITS must be evenly divisible by eight.
@@ -860,3 +1051,45 @@
       (unless (= shift initial)
         (vector-push-extend byte byte-buffer))))
   (values))
+
+(declaim (ftype (sfunction ((and byte-buffer (not simple-array)))
+                           (or (simple-array (unsigned-byte 8) 1)
+                               (simple-array (signed-byte 8) 1)))
+                compress)
+         (ftype (sfunction ((or (simple-array (unsigned-byte 8) 1)
+                                (simple-array (signed-byte 8) 1)))
+                           (simple-array (unsigned-byte 8) 1))
+                decompress))
+
+#-(and sb-core-compression (not sb-xc-host))
+(progn
+  (defun compress (vector)
+    #-sb-xc-host (%shrink-vector (%array-data vector) (length vector))
+    #+sb-xc-host (coerce vector '(simple-array (unsigned-byte 8) 1)))
+
+  (defun decompress (vector)
+    vector))
+
+#+(and sb-core-compression (not sb-xc-host))
+(progn
+  (defun compress (vector)
+    (with-alien ((compress-vector (function int (* char) size-t) :extern "compress_vector"))
+      (let* ((data (truly-the (simple-array * (*)) (%array-data vector))))
+        (with-pinned-objects (data)
+          (alien-funcall compress-vector (int-sap (get-lisp-obj-address data)) (length vector)))
+        data)))
+
+  (defun decompress (vector)
+    (if (typep vector '(simple-array (signed-byte 8) (*)))
+        (with-alien ((decompress-vector (function (* unsigned-char) (* char) (* size-t)) :extern "decompress_vector")
+                     (size size-t))
+          (let* ((pointer
+                   (with-pinned-objects (vector)
+                     (alien-funcall decompress-vector (int-sap (get-lisp-obj-address vector)) (addr size))))
+                 (length size)
+                 (new (make-array length :element-type '(unsigned-byte 8))))
+            (loop for i below length
+                  do (setf (aref new i) (deref pointer i)))
+            (free-alien pointer)
+            new))
+        vector)))
