@@ -325,7 +325,8 @@
 
 (defknown atan
   (number &optional real) irrational
-  (movable foldable unsafely-flushable recursive))
+  (movable foldable unsafely-flushable recursive)
+  :call-type-deriver #'atan-call-type-deriver)
 
 (defknown (tan sinh cosh tanh asinh)
   (number) irrational (movable foldable flushable recursive))
@@ -357,6 +358,9 @@
   (real &optional real) (values integer real)
   (movable foldable flushable recursive))
 
+(defknown (sb-kernel::truncate1 sb-kernel::floor1 sb-kernel::ceiling1) (real real) integer
+  (movable foldable flushable recursive no-verify-arg-count))
+
 (defknown unary-truncate (real) (values integer real)
   (movable foldable flushable no-verify-arg-count))
 
@@ -1016,12 +1020,11 @@
 (defknown sb-impl::|Vector| (&rest t) simple-vector (movable flushable))
 
 ;;; All but last must be of type LIST, but there seems to be no way to
-;;; express that in this syntax.
+;;; express that in this syntax, append-call-type-deriver does that.
 (defknown append (&rest t) t (flushable)
   :call-type-deriver #'append-call-type-deriver)
-(defknown sb-impl::append2 (list t) t
-  (flushable no-verify-arg-count)
-  :call-type-deriver #'append-call-type-deriver)
+(defknown sb-impl::append2 (proper-list t) t
+  (flushable no-verify-arg-count))
 
 (defknown copy-list (proper-or-dotted-list) list (flushable)
   :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t))
@@ -1031,7 +1034,8 @@
 (defknown copy-tree (t) t (flushable recursive))
 (defknown revappend (proper-list t) t (flushable))
 
-(defknown nconc (&rest (modifying t :butlast t)) t ())
+(defknown nconc (&rest (modifying t :butlast t)) t ()
+  :call-type-deriver #'nconc-call-type-deriver)
 
 (defknown nreconc ((modifying list) t) t (important-result))
 (defknown butlast (proper-or-dotted-list &optional unsigned-byte) list (flushable))
@@ -1145,8 +1149,8 @@
      &key (:key (function-designator ((nth-arg 1 :sequence t)))))
     list (foldable flushable call))
 
-(defknown (memq assq) (t proper-list) list (foldable flushable))
-(defknown (delq delq1) (t (modifying list)) list (flushable))
+(defknown (memq assq) (t proper-list) list (foldable flushable no-verify-arg-count))
+(defknown (delq delq1) (t (modifying list)) list (flushable no-verify-arg-count))
 
 ;;;; from the "Hash Tables" chapter:
 
@@ -2317,14 +2321,13 @@
            sb-thread::call-with-recursive-system-lock)
     ((function ()) t) *))
 
-#+round-float
 (progn
-  (defknown round-double (double-float #1=(member :round :floor :ceiling :truncate))
+  (defknown round-double (double-float #1=(member #+round-float :round :floor :ceiling :truncate))
       double-float
-      (foldable flushable movable always-translatable))
+      (foldable flushable movable #+round-float always-translatable))
 
   (defknown round-single (single-float #1#) single-float
-      (foldable flushable movable always-translatable)))
+      (foldable flushable movable #+round-float always-translatable)))
 
 (defknown (overflow* overflow+ overflow-
            overflow-ash)
