@@ -27,7 +27,8 @@
                           character-widetag)))
       (structure-object
        (if (eq val sb-lockless:+tail+)
-           (inst addi y null-tn (- sb-vm::lockfree-list-tail-value sb-vm:nil-value))
+           (inst addi y null-tn (- lockfree-list-tail-value-offset
+                                   nil-value-offset))
            (bug "immediate structure-object ~S" val))))))
 
 (define-move-fun (load-number 1) (vop x y)
@@ -50,12 +51,7 @@
       (short-immediate
        (loadw y code-tn (tn-offset x) other-pointer-lowtag))
       (u+i-immediate
-       (multiple-value-bind (u i)
-           (u-and-i-inst-immediate offset)
-         ;; Should be GC safe.
-         (inst lui y u)
-         (inst add lip-tn code-tn y)
-         (inst #-64-bit lw #+64-bit ld y lip-tn i))))))
+       (inst load-far-constant y x)))))
 
 (define-move-fun (load-stack 5) (vop x y)
   ((control-stack) (any-reg descriptor-reg))
@@ -164,16 +160,15 @@
   (:note "integer to untagged word coercion")
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 3
-    (let ((done (gen-label)))
-      (inst andi temp x fixnum-tag-mask)
-      (sc-case y
-        (signed-reg
-         (inst srai y x n-fixnum-tag-bits))
-        (unsigned-reg
-         (inst srli y x n-fixnum-tag-bits)))
-      (inst beq temp zero-tn done)
-      (loadw y x bignum-digits-offset other-pointer-lowtag)
-      (emit-label done))))
+    (inst andi temp x fixnum-tag-mask)
+    (sc-case y
+      (signed-reg
+       (inst srai y x n-fixnum-tag-bits))
+      (unsigned-reg
+       (inst srli y x n-fixnum-tag-bits)))
+    (inst beq temp zero-tn done)
+    (loadw y x bignum-digits-offset other-pointer-lowtag)
+    DONE))
 
 (define-move-vop move-to-word/integer :move
   (descriptor-reg) (signed-reg unsigned-reg))
