@@ -19,13 +19,13 @@
 #include <stdio.h>
 #include <stdbool.h>
 #include <signal.h>
+#include <stdlib.h>
 #include "genesis/sbcl.h"
 #include "runtime.h"
 #include "globals.h"
 #include "os.h"
 #include "interrupt.h"
 #include "lispregs.h"
-#include <wchar.h>
 #include "arch.h"
 #include "genesis/compiled-debug-info.h"
 #include "genesis/hash-table.h"
@@ -56,36 +56,39 @@ debug_function_name_from_pc (struct code* code, void *pc)
 
     uword_t offset = (char*)pc - code_text_start(code);
 
-    struct vector *v;
-    int i, len;
-
-    v = VECTOR(di->fun_map);
-
-    len = vector_len(v);
-
+    struct vector *v = VECTOR(di->fun_map);
+    sword_t len = vector_len(v);
     unsigned char *map = (unsigned char*)v->data;
 
+#ifdef LISP_FEATURE_SB_CORE_COMPRESSION
+    int compressed = widetag_of(&v->header) == SIMPLE_ARRAY_SIGNED_BYTE_8_WIDETAG;
+    if (compressed)
+        map = decompress_vector(di->fun_map, (size_t*)&len);
+#endif
+
     uword_t code_start_pc = 0;
     __attribute__((unused)) uword_t start_pc = 0;
     uword_t elsewhere_pc = 0;
     uword_t first_elsewhere_pc = 0;
     lispobj last_name = 0;
-    lispobj name;
+    lispobj name = di->name;
 
-    i = 0;
+    int i = 0;
     while (i < len) {
         unsigned char options = map[i++];
         unsigned char flags = map[i++];
         bool vars_p = flags & PACKED_DEBUG_FUN_VARIABLES_BIT;
         bool blocks_p = flags & PACKED_DEBUG_FUN_BLOCKS_BIT;
 
-        if (is_lisp_pointer(di->contexts) &&
-            widetag_of(native_pointer(di->contexts)) == SIMPLE_VECTOR_WIDETAG) {
-            struct vector *contexts = VECTOR(di->contexts);
-            name = contexts->data[read_var_integer(map, &i)];
-        } else {
-            read_var_integer(map, &i);
-            name = di->contexts;
+        if (!(flags & PACKED_DEBUG_FUN_PREVIOUS_NAME)) {
+            if (is_lisp_pointer(di->contexts) &&
+                widetag_of(native_pointer(di->contexts)) == SIMPLE_VECTOR_WIDETAG) {
+                struct vector *contexts = VECTOR(di->contexts);
+                name = contexts->data[read_var_integer(map, &i)];
+            } else {
+                read_var_integer(map, &i);
+                name = di->contexts;
+            }
         }
 
         if (vars_p) {
@@ -153,20 +156,26 @@ debug_function_name_from_pc (struct code* code, void *pc)
         if (first_elsewhere_pc) {
             if ((offset >= first_elsewhere_pc &&
                  offset < elsewhere_pc))
-                return last_name;
+                goto done;
         } else
             first_elsewhere_pc = elsewhere_pc;
 
         if (offset < code_start_pc)
-            return last_name;
+            goto done;
 
         last_name = name;
     }
 
     if (i == len)
-        return last_name;
+        goto done;
     else
         lose("failed to parse debug function name");
+  done:
+#ifdef LISP_FEATURE_SB_CORE_COMPRESSION
+    if (compressed)
+        free(map);
+#endif
+    return last_name;
 }
 
 static void
@@ -278,12 +287,13 @@ static void __attribute__((unused))
 print_entry_points (struct code *code, FILE *f)
 {
     int n_funs = code_n_funs(code);
+    struct compiled_debug_info* cdi = (void*)native_pointer(barrier_load(&code->debug_info));
     for_each_simple_fun(index, fun, code, 0, {
         if (widetag_of(&fun->header) != SIMPLE_FUN_WIDETAG) {
             fprintf(f, "%p: bogus function entry", fun);
             return;
         }
-        print_entry_name(barrier_load(&code->constants[CODE_SLOTS_PER_SIMPLE_FUN*index]),
+        print_entry_name(barrier_load(CODE_SLOTS_PER_SIMPLE_FUN*index + &cdi->rest),
                          f);
         if ((index + 1) < n_funs) fprintf(f, ", ");
     });
