@@ -98,12 +98,6 @@ extern void enliven_immobile_obj(lispobj*,int);
 //                                   v    v
 //                       0xzzzzzzzz GGzzzzww
 //         arbitrary data  --------   ---- length in words
-//
-// An an exception to the above, FDEFNs omit the length:
-//                       0xzzzzzzzz zzzzGGww
-//         arbitrary data  -------- ----
-// so that there are 6 consecutive bytes of arbitrary data.
-// The length of an FDEFN is implicitly fixed at 4 words.
 
 // There is a hard constraint on NUM_GENERATIONS, which is currently 8.
 // (0..5=normal, 6=pseudostatic, 7=scratch)
@@ -121,28 +115,19 @@ static inline int immobile_obj_gen_bits(lispobj* obj) // native pointer
     // When debugging, assert that we're called only on a headered object
     // whose header contains a generation byte.
     gc_dcheck(!embedded_obj_p(widetag_of(obj)));
-    char gen;
-    switch (widetag_of(obj)) {
-    default:
-        gen = ((generation_index_t*)obj)[3]; break;
-    case FDEFN_WIDETAG:
-        gen = ((generation_index_t*)obj)[1]; break;
-    }
-    return gen & 0x1F;
+    return ((generation_index_t*)obj)[3] & 0x1F;
 }
 // Turn a grey node black.
 static inline void set_visited(lispobj* obj)
 {
     gc_dcheck(widetag_of(obj) != SIMPLE_FUN_WIDETAG);
     gc_dcheck(immobile_obj_gen_bits(obj) == new_space);
-    int byte = widetag_of(obj) == FDEFN_WIDETAG ? 1 : 3;
-    ((generation_index_t*)obj)[byte] |= IMMOBILE_OBJ_VISITED_FLAG;
+    ((generation_index_t*)obj)[3] |= IMMOBILE_OBJ_VISITED_FLAG;
 }
 static inline void assign_generation(lispobj* obj, generation_index_t gen)
 {
     gc_dcheck(widetag_of(obj) != SIMPLE_FUN_WIDETAG);
-    int byte = widetag_of(obj) == FDEFN_WIDETAG ? 1 : 3;
-    generation_index_t* ptr = (generation_index_t*)obj + byte;
+    generation_index_t* ptr = (generation_index_t*)obj + 3;
     // Clear the VISITED flag, assign a new generation, preserving the three
     // high bits which include the OBJ_WRITTEN flag as well as two
     // opaque flag bits for use by Lisp.
