@@ -146,21 +146,6 @@ struct page {
     generation_index_t gen;
 };
 extern struct page *page_table;
-extern page_index_t page_table_pages;
-
-/* Find the page index within the page_table for the given
- * address. Return -1 on failure. */
-static inline page_index_t find_page_index(void *addr)
-{
-    if (addr >= (void*)DYNAMIC_SPACE_START) {
-        page_index_t index = ((uintptr_t)addr -
-                              (uintptr_t)DYNAMIC_SPACE_START) / GENCGC_PAGE_BYTES;
-        if (index < page_table_pages)
-            return (index);
-    }
-    return (-1);
-}
-extern char *page_address(page_index_t);
 
 /* New objects are allocated to PAGE_TYPE_MIXED or PAGE_TYPE_CONS */
 /* If you change these constants, then possibly also change the following
@@ -344,9 +329,9 @@ extern void gc_close_collector_regions(int);
 /* The various sorts of pointer swizzling in SBCL. */
 enum source {
   SOURCE_NORMAL,
+  SOURCE_LINKAGE_CELL,
   SOURCE_ZERO_TAG,              /* code, lflist */
   SOURCE_CLOSURE,
-  SOURCE_SYMBOL_NAME,
   SOURCE_FDEFN_RAW
 };
 
@@ -557,7 +542,6 @@ extern char * gc_logfile;
 extern void log_generation_stats(char *logfile, char *header);
 extern void print_generation_stats(void);
 extern double generation_average_age(generation_index_t);
-#define PAGE_INDEX_FMT PRIdPTR
 static inline os_vm_size_t npage_bytes(page_index_t npages)
 {
     gc_assert(npages>=0);
@@ -600,7 +584,7 @@ set_page_scan_start_offset(page_index_t index, os_vm_size_t offset)
 {
     // If the offset is nonzero and page-aligned
     unsigned int lsb = offset !=0 && IS_ALIGNED(offset, GENCGC_PAGE_BYTES);
-    os_vm_size_t scaled = (offset >> (lsb ? GENCGC_CARD_SHIFT-1 : WORD_SHIFT)) | lsb;
+    os_vm_size_t scaled = (offset >> (lsb ? GENCGC_PAGE_SHIFT-1 : WORD_SHIFT)) | lsb;
     if (scaled > SCAN_START_OFS_MAX) {
         // Assert that if offset exceed the max representable value,
         // then it is a page-aligned offset, not a cons-aligned offset.
@@ -623,7 +607,7 @@ static os_vm_size_t scan_start_offset_iterated(page_index_t index)
         offset = page_table[lookback_page].scan_start_offset_;
         tot_offset_in_pages += offset >> 1;
     } while (offset == SCAN_START_OFS_MAX);
-    return (os_vm_size_t)tot_offset_in_pages << GENCGC_CARD_SHIFT;
+    return (os_vm_size_t)tot_offset_in_pages << GENCGC_PAGE_SHIFT;
 }
 
 static os_vm_size_t  __attribute__((unused)) page_scan_start_offset(page_index_t index)
@@ -631,7 +615,7 @@ static os_vm_size_t  __attribute__((unused)) page_scan_start_offset(page_index_t
     return page_table[index].scan_start_offset_ != SCAN_START_OFS_MAX
         ? (os_vm_size_t)(page_table[index].scan_start_offset_ & ~1)
           << ((page_table[index].scan_start_offset_ & 1) ?
-              (GENCGC_CARD_SHIFT-1) : WORD_SHIFT)
+              (GENCGC_PAGE_SHIFT-1) : WORD_SHIFT)
         : scan_start_offset_iterated(index);
 }
 
@@ -776,8 +760,6 @@ extern struct fixedobj_page *fixedobj_pages;
 #define fixedobj_page_obj_align(i) (fixedobj_pages[i].attr.parts.obj_align<<WORD_SHIFT)
 #endif
 
-extern page_index_t next_free_page;
-
 extern uword_t
 walk_generation(uword_t (*proc)(lispobj*,lispobj*,uword_t),
                 generation_index_t generation, uword_t extra);
